#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "aamodels.h"
#include "insert_freqs.h"
#include "global.h"
#include "tree.h"
#include "evolve.h"
#include "model.h"
#include "gamma.h"
#include "twister.h"
#include "treefile.h"
#include "random.h"

/*CS:Definitions*/
#define INSERT 0
#define DELETE 1
#define NO_ACTION 2
#define INVARIABLE 1
#define NO_INDEL 2
#define INVAR_AND_NOINDEL 3
int numTaxa, numSites, maxPartitions, numPartitions, fileFormat;
static double *NeighborRates;
char *global_alignment_action;
int  *global_alignment_fromAnc;
int   global_alignment_end;
/**/

double gammaShape, proportionInvariable;
int numCats, rateHetero, invariableSites,/*CS:*/tree_indel_buffer,indel_fill_aas;
double catRate[MAX_RATE_CATS];
double freqRate[MAX_RATE_CATS];

static double *matrix[MAX_RATE_CATS];
static double *vector;

static double *gammaRates;
static short *categories;
static short *invariable;
static double *siteRates;

int insert_debug = 0;
int delete_debug = 0;

/* prototypes */

char SetState(double *P);
short IsInvariable();
void SetSequence(char *seq, char *source, int inFromSite, int inNumSites);
void SetNucSequence(char *seq, char *source, int inFromSite, int inNumSites);
void SetAASequence(char *seq, char *source, int inFromSite, int inNumSites);
void RandomSequence(char *seq, int inNumSites);
void MutateSequence(TNode *des, char *seq, int inFromSite, int inNumSites, double len);
void SetSiteRates(TNode *des, int inFromSite, int inNumSites, double inScale);
void EvolveNode(/*CS:*/TTree *tree, TNode *anc, TNode *des, int inFromSite, int inNumSites, double scale, int *P);
void WritePhylipFormat(FILE *fv, TTree **treeSet, int *partitionLengths);
void WriteNexusFormat(FILE *fv, int treeNo, int datasetNo, TTree **treeSet, int *partitionLengths);
void WriteAncestralSequencesNode(FILE *fv, TTree *tree, int *nodeNo, TNode *des);
/*CS:Prototypes*/
int indel_buffer(TTree *tree);
void InDel(TTree *tree, TNode *des, double len);
int Find_Indel_Size(double *dist,int max_indel);
void Insert(TNode *des, int indel_size);
void Delete(TNode *des, int indel_size);
int  Find_Start_Location(TNode *des, int *indel_size, int final_aa, int action);
int  Find_Action(double len, double P_insert_, double P_delete_);
int  isAnc(TNode *thisNode, int anc);
int  isSpot(int *pos, TNode *curr_taxa);
void Create_indelRates();
void insertFillSequence(char *seq, int site, int inNumSites);
char indelSetState(double *P);
void Create_Global_Arrays(TTree *tree);
void Print_Globals();
//void WriteMA(FILE *fv, TTree *tree);

/* functions */

void CreateRates()
{
int i;

    //fprintf(stderr,"CreateRates %d\n",invariableSites);
    if (rateHetero==GammaRates) gammaRates=CAllocMem(numSites*sizeof(double), "gammaRates", "CreateRates", 0);
    else if (rateHetero==DiscreteGammaRates) categories=CAllocMem(numSites*sizeof(short), "categories", "CreateRates", 0);

    if (invariableSites) invariable=CAllocMem((numSites)*sizeof(short), "invariable", "CreateRates", 0);

    siteRates=CAllocMem(numSites*sizeof(double), "siteRates", "CreateRates", 0);

    for (i = 0; i < MAX_RATE_CATS; i++) {
	matrix[i] = CAllocMem(numStates*numStates*sizeof(double), "matrix", "CreateRates", 0);
    }
    NeighborRates = (double*)malloc(400*sizeof(double));
    Set_Neighbor_Rates(NeighborRates,indel_fill_aas);
    vector = CAllocMem(numStates*sizeof(double), "vector", "CreateRates", 0);
}

void Create_indelRates(TNode *node)
{
int i;

    if (rateHetero==GammaRates) 
	node->gammaRates=CAllocMem((numSites+tree_indel_buffer+1)*sizeof(double), "gammaRates", "CreateRates", 0);
    else if (rateHetero==DiscreteGammaRates) 
	node->categories=CAllocMem((numSites+tree_indel_buffer+1)*sizeof(short), "categories", "CreateRates", 0);

    if (invariableSites) {
	node->invariable=CAllocMem((numSites+tree_indel_buffer+1)*sizeof(short), "invariable", "CreateRates", 0);
    	for(i=0; i<numSites+tree_indel_buffer+1; i++) invariable[i] = 0;
    }

    node->siteRates=CAllocMem((numSites+tree_indel_buffer+1)*sizeof(double), "siteRates", "CreateRates", 0);
}

void Create_Global_Arrays(TTree *tree) {
    int i;

    global_alignment_action=(char*)malloc(2*(2*numSites+(int)(numSites*0.0224*tree->maxIndel*tree->numNodes))*sizeof(char));
    global_alignment_fromAnc=(int*)malloc(2*(2*numSites+(int)(numSites*0.0224*tree->maxIndel*tree->numNodes))*sizeof(int));

    //fprintf(stderr,"Allocated %d sites for the global arrays.\n",2*numSites+(int)(numSites*0.0224*tree->maxIndel*tree->numNodes));

    for(i=0; i<=numSites; i++) {
        global_alignment_action[i] = 'i';
        global_alignment_fromAnc[i] = -1;
    } 

    global_alignment_end = numSites+1; 
}

void CreateSequences(TTree *tree, int inNumSites)
{
    TNode *P;

    tree_indel_buffer=indel_buffer(tree);

    P=tree->nodeList;
    while (P!=NULL) {
	if (P->sequence != NULL) free(P->sequence);	
	if (tree->maxIndel!=0) {
	    P->sequence=CAllocMem(inNumSites+tree_indel_buffer+1, "sequences", "CreateSequences", 0);
	    //fprintf(stderr,"total sites=%d\t indel buffer=%d\n",inNumSites+tree_indel_buffer,tree_indel_buffer);
	} else {
	    P->sequence=CAllocMem(inNumSites+1, "sequences", "CreateSequences", 0);
	}
	P=P->next;
    }
}

/*CS: This function calculates the expected size of the indel buffer needed. */
int indel_buffer(TTree *tree) {
    int buffer_size;
    double rate;
    double tree_depth;
    double E_indelSize_=0;
    int i;

    tree_depth=log10(tree->numNodes)/log10(2);

    if(tree->P_ins_ == 0 && tree->P_del_ == 0) {
	buffer_size=(int)((double)numSites*0.0224*(double)tree->maxIndel*0.5*tree_depth);
    } else {
	if(tree->P_del_==0) { rate = tree->P_ins_ * (double)numSites; }
	else { rate=tree->P_ins_/tree->P_del_; }
	for(i=1; i<=tree->maxIndel; i++) { E_indelSize_+=(double)i*tree->insert_lengthDistribution[i]; }
	buffer_size=(int)(rate*E_indelSize_*tree_depth);
    }

    buffer_size+=numSites/2;	/* Just a little extra for caution. */

    return 2*buffer_size;
}
/**/

void SetCategories()
{
    int i;
    double sumRates;
	
    if (rateHetero==CodonRates) {
	sumRates=catRate[0]+catRate[1]+catRate[2];
	if (sumRates!=3.0) {
	    catRate[0]*=3.0/sumRates;
	    catRate[1]*=3.0/sumRates;
	    catRate[2]*=3.0/sumRates;
	}
    } else if (rateHetero==GammaRates) {
	for (i=0; i<numSites; i++)
	    gammaRates[i]=rndgamma(gammaShape) / gammaShape;
    } else if (rateHetero==DiscreteGammaRates) {
	DiscreteGamma(freqRate, catRate, gammaShape, gammaShape, numCats, 0);
	for (i=0; i<numSites; i++) categories[i]=(int)(rndu()*numCats);
    }

    if (invariableSites) {
	for (i=0; i<numSites; i++) invariable[i] = IsInvariable();
    } 
}

char SetState(double *P)
{
    char j;
    double r;
	
    r=rndu();
    for (j=0; r>(*P) && j<numStates; j++) P++;
    return j;
}

short IsInvariable()
{
    double r;

    r=rndu();
    if (r < proportionInvariable) return INVARIABLE;
    else return 0;
}


void RandomSequence(char *seq, int inNumSites)
{
    int i;
    char *P;
	
    P=seq;
    for (i=0; i<inNumSites; i++) {
	*P=SetState(addFreq);
	P++;
    }
}

void insertFillSequence(char *seq, int site, int inNumSites)
{
    int i,j;
    int iFS_debug = 0;
    char *P;
    double *bayes_probabilities_for_next,r;

    P=seq+site;
    /* Bayes Rule: P(nextAA=j|thisAA=i) = ( (P(thisAA=i|nextAA=j) * freqAA[nextAA]) / freqAA[thisAA]) */
    /* Need to calculate each of the relative probabilities (20 of them) based on Bayes above, and then
     * make a flip, and place that AA in as the next.  The first AA is in P=seq-1, unless at the first
     * position of the sequence, then we do not know, and will pick randomly. */
    if(site==0) {
        /* In this case, the previous amino acid is unknown, so generate a random flip to start. */
        *P=indelSetState(addFreq);
        P++;	// Set first state. //
        inNumSites--;	/* One amino acid generated, need to generate one less. */
    }
    for(i=0; i<inNumSites; i++) {
        bayes_probabilities_for_next=(double*)malloc(20*sizeof(double));
	if(iFS_debug) {
	    fprintf(stderr,"1: bayes_probabilities_for_next[0]=(aaFreq[0](%f) * NeighborRates[(int)*(P-1)*20(%d)] (%f))/(aaFreq[*(P-1) (%d)] (%f)",aaFreq[0],(int)*(P-1)*20,NeighborRates[(int)*(P-1)*20],*(P-1),aaFreq[*(P-1)]);
	    fprintf(stderr," = %f\n", (aaFreq[0]*NeighborRates[(int)*(P-1)*20])/(aaFreq[*(P-1)]));
	}
	bayes_probabilities_for_next[0]=(aaFreq[0]*NeighborRates[(int)*(P-1)*20])/(aaFreq[*(P-1)]);
        for(j=1/*0*/;j<20;j++) {
	    bayes_probabilities_for_next[j]=bayes_probabilities_for_next[j-1];
	    bayes_probabilities_for_next[j]+=(aaFreq[j]*NeighborRates[(int)*(P-1)*20+j])/(aaFreq[*(P-1)]);
	    if(iFS_debug) {
	 	fprintf(stderr,"(%lf * %lf) / %lf = ",aaFreq[j],NeighborRates[(int)*(P-1)*20+j],aaFreq[*(P-1)]);
	    	fprintf(stderr,"%lf",bayes_probabilities_for_next[j]);
	    	fprintf(stderr,"\tNeighborRates[%d*20+%d(%d)]\n",*(P-1),j,(int)*(P-1)*20+j);
	    }
	}

	/* Probability vectors now filled out. */
	r=rndu();
	if(iFS_debug) fprintf(stderr,"random_num=%lf\n",r);
	r*=bayes_probabilities_for_next[19];
	j=0;
	if(iFS_debug) fprintf(stderr,"Mult by bayes prob. vec.=%lf\n",r);
	while(r>bayes_probabilities_for_next[j]) { 
	    if(iFS_debug) fprintf(stderr,"Rand=%lf,\tb_p_f_n=%lf\n",r,bayes_probabilities_for_next[j]);
	    j++; 
	}
	*P=j;
	P++;
    }

}

char indelSetState(double *P)
{
    char j;
    double r;

    r=rndu();
    for (j=0; r>(*P) && j<numStates; j++) P++;
    return j;
}


void SetSequence(char *seq, char *source, int inFromSite, int inNumSites)
{
    int i, j;
    char *P, *Q;
	
    P=seq;
    Q=source + inFromSite;
    for (i=0; i<inNumSites; i++) {
	for (j=0; j < numStates; j++) {
	    if (*Q == stateCharacters[j]) break;
	}
	if (j == numStates) {
	    fprintf(stderr, "Bad state in ancestoral sequence\n");
	    exit(0);
	}
	*P = j;
	P++;
	Q++;
    }
}

void MutateSequence(TNode *des, char *seq, int inFromSite, int inNumSites, double len)
{
    int i, j, cat;
    double *Q;
    short *R;
    short *S;
    char *P;
	
    P=seq;
	
    /* This is done quite long-windedly to keep the speed up */
    switch (rateHetero) {
	case GammaRates:
	    Q=gammaRates + inFromSite;
			
	    if (invariableSites) {
		if(indel_flag) S = des->invariable + inFromSite;
		else S = invariable + inFromSite;
		for (i=0; i<inNumSites; i++) {
		    if ((*S)!=INVARIABLE && (*S)!=INVAR_AND_NOINDEL) {
			SetVector(vector, *P, (*Q)*len);
			*P=SetState(vector);
		    }
		    P++;
		    Q++;
		    S++;
		}
	    } else {
		for (i=0; i<inNumSites; i++) {
		    SetVector(vector, *P, (*Q)*len);
		    *P=SetState(vector);
		    P++;
		    Q++;
		}
	    }
	    break;

	case DiscreteGammaRates:
	    for (i=0; i<numCats; i++)
	    SetMatrix(matrix[i], catRate[i]*len);
			
	    R=categories + inFromSite;
	    if (invariableSites) {
		if(indel_flag) S = des->invariable + inFromSite;
		else S = invariable + inFromSite;
		for (i=0; i<inNumSites; i++) {
		    if ((*S)!=INVARIABLE && (*S)!=INVAR_AND_NOINDEL) 
			*P=SetState(matrix[*R]+((*P) * numStates));
		    P++;
		    R++;
		    S++;
		}
	    } else {
		for (i=0; i<inNumSites; i++) {
		    *P=SetState(matrix[*R]+((*P) * numStates));
		    P++;
		    R++;
	    	}
	    }
	    break;

	case CodonRates:
	    for (i=0; i<numCats; i++) SetMatrix(matrix[i], catRate[i]*len);
	    j = inFromSite;
	    for (i=0; i< inNumSites; i++) {
		cat=j%3;
		*P=SetState(matrix[cat]+((*P) * numStates));
		P++;
		j++;
	    }
	    break;

	case NoRates:
	    SetMatrix(matrix[0], len);
	    if (invariableSites) {
		if(indel_flag) S = des->invariable + inFromSite;
		else S = invariable + inFromSite;
		for (i=0; i<inNumSites; i++) {
		    if ((*S)!=INVARIABLE && (*S)!=INVAR_AND_NOINDEL) 
			*P=SetState(matrix[0]+((*P) * numStates));
			P++;
			S++;
		}
	    } else {
		for (i=0; i<inNumSites; i++) {
		    *P=SetState(matrix[0]+((*P) * numStates));
		    P++;
		}
	    }
	    break;
    }
}


void SetSiteRates(TNode *des, int inFromSite, int inNumSites, double inScale)
{
    int i, j, cat;
    double *Q;
    short *R;
    short *S;
    double *P;
	
    if(indel_flag) { 
	if(invariableSites) P=des->siteRates + inFromSite;
    } else P=siteRates + inFromSite;

    /* This is done quite long-windedly to keep the speed up */
    switch (rateHetero) {
    case GammaRates:
	Q=gammaRates + inFromSite;
			
	if (invariableSites) {
	    if(indel_flag) S = des->invariable + inFromSite;
	    else S = invariable + inFromSite;
	    for (i=0; i<inNumSites; i++) {
		if ((*S)!=INVARIABLE && (*S)!=INVAR_AND_NOINDEL) *P = (*Q) * inScale;
		else *P = 0.0;
		P++;
		Q++;
		S++;
	    }
	} else {
	    for (i=0; i<inNumSites; i++) {
		*P = (*Q) * inScale;
		P++;
		Q++;
	    }
	}
	break;

    case DiscreteGammaRates:
	R=categories + inFromSite;

	if (invariableSites) {
	    if(indel_flag) S = des->invariable + inFromSite;
	    else S = invariable + inFromSite;
	    for (i=0; i<inNumSites; i++) {
		if ((*S)!=INVARIABLE && (*S)!=INVAR_AND_NOINDEL) *P = catRate[*R] * inScale;
		else *P = 0.0;
		P++;
		R++;
		S++;
	    }
	} else {
	    for (i=0; i<inNumSites; i++) {
 		*P = catRate[*R] * inScale;
		P++;
		R++;
	    }
	}
	break;

    case CodonRates:
	j = inFromSite;
	for (i=0; i< inNumSites; i++) {
	    cat=j%3;
	    *P = catRate[cat] * inScale;
	    P++;
	    j++;
	}
	break;

    case NoRates:
	if (invariableSites) {
	    if(indel_flag) S = des->invariable + inFromSite;
	    else S = invariable + inFromSite;
	    for (i=0; i<inNumSites; i++) {
		if ((*S)!=INVARIABLE && (*S)!=INVAR_AND_NOINDEL) *P = inScale;
		else *P = 0.0;
		P++;
		S++;
	    }
	} else {
	    for (i=0; i<inNumSites; i++) {
		*P = inScale;
		P++;
	    }
	}
	break;
    }
}

/*CS: This is the main insertion and deletion routine, made to find the indel sites. */
void InDel(TTree *tree, TNode  *des, double len) {
    int action = INSERT;
    int indel_size;
    int i;
    int number_of_noIndel_positions=0;
    int possible_pos=des->indels;
    int step = 0;

    if(invariableSites) {
	if(insert_debug || delete_debug) fprintf(stderr,"Invariable: ");
	for(i=0; i<numSites+des->indels; i++) {
	    if(insert_debug || delete_debug) fprintf(stderr,"%d",des->invariable[i]);
	    if(des->invariable[i]) number_of_noIndel_positions++;
	}
	if(insert_debug || delete_debug) fprintf(stderr,"\n");
    }

    /* This for loop does not choose site-specific indel sites, i.e. does not check for an indel between each
     * position in the sequence.  Rather, since we may be dealing with invariable and noIndel sites, we cannot
     * simply randomly choose the site of an indel.  For insertions, we need to make sure that we do not make
     * an insertion between two sites where des->invariant!=0||1 in consecutive positions.  For a deletion, we need
     * to make sure that we do not delete a position that is invariant.  Furthermore, we are limited in the number
     * of positions where we can make a deletion, since starting a deletion anywhere within maxIndel sites from an
     * invariant site will delete the invariant site. */
    for(i=0; i<numSites+possible_pos-number_of_noIndel_positions; i++) {
	action = Find_Action(len,tree->P_ins_,tree->P_del_);  /* determines if a delete, insert, or nothing happens. */
	if( (insert_debug && (action == INSERT || action == DELETE)) &&
	    (delete_debug && (action == INSERT || action == DELETE)) )
	    fprintf(stderr,"action = %s\n",((action)?"deletion":"insertion")); 
	if(action == INSERT || action == DELETE) {
	    if(action == INSERT) {
	        indel_size=Find_Indel_Size(tree->insert_lengthDistribution,tree->maxIndel);
	        if(insert_debug || delete_debug) 
		    fprintf(stderr,"%s size = %d\t (found iteration %d)\n",((action)?"deletion":"insertion"),indel_size,i);
		/* Just make sure not to put insertion in between consecutive invariable positions. */
		Insert(des,indel_size);
		des->indels+=indel_size;
	    } else {
		/* More difficult finding a spot to put deletion, maybe impossible... */
	        indel_size=Find_Indel_Size(tree->delete_lengthDistribution,tree->maxIndel);
	        if(insert_debug || delete_debug) 
		    fprintf(stderr,"%s size = %d\t (found iteration %d)\n",((action)?"deletion":"insertion"),indel_size,i);
		if((numSites+des->indels)-indel_size <= 0) {
		    if(insert_debug || delete_debug) 
			fprintf(stderr, "Deletion of size %d occurred on sequence of size %d\n",indel_size,numSites+des->indels);
		}	/* Make a defunct protein, cannot be done. */
		else {
		    Delete(des,indel_size);
		    des->indels-=indel_size;
		}
	    }
	}
    }
}

void Insert(TNode *des, int indel_size) {
    int start_aa;
    int final_aa;
    int i,j;
    int prev_pos=0,indel_aa_in_globals=0,num_skip=0;
    int *position_in_globals,*num_indels_in_globals;
    int step = 0;

    position_in_globals= (int*)malloc(sizeof(int));
    *position_in_globals= 0;
    num_indels_in_globals= (int*)malloc(sizeof(int));
    *num_indels_in_globals= 0;

    final_aa=numSites+des->indels+indel_size; /* indel_size is added into des->indels after this function is called */
    start_aa=Find_Start_Location(des,&indel_size,numSites+des->indels,INSERT);

    if(insert_debug) fprintf(stderr,"INSERT: size = %d,\tspot = %d,\tfinal_aa = %d\tdes->tipNo = %d\n",indel_size,start_aa,final_aa,des->tipNo);

    /* Sanity check, core dump saver. */
    if(start_aa+indel_size > final_aa) {
	fprintf(stderr,"start_aa=%d\t+\tindel_size=%d\t>\tfinal_aa=%d\n",start_aa,indel_size,final_aa);
	exit(0);
    }

    /* Here also, I need to find the correct spot in the global_* arrays, and place the 
     * insertion symbol, as well as the TNode* tipNo.
     * First, get to the indel_aa spot in the global arrays.  Because indel_aa is the position
     * before the insertion, we need to skip over indel_aa.  OH, not true.  Stay before indel_aa. */
    while(indel_aa_in_globals <= start_aa) { 
        if(*position_in_globals > global_alignment_end) { 
            fprintf(stderr,"INSERT: Over g_a_end.  indel_aa_in_globals = %d,\tstart_aa = %d\n", indel_aa_in_globals,start_aa);
            exit(0); 
        } 
        prev_pos = *position_in_globals; 
        if(isSpot(position_in_globals,des)) { indel_aa_in_globals++; } 
        *position_in_globals += 1; 
        //if(prev_pos == *position_in_globals) { printf("prev_pos = p_i_g, %d\n",prev_pos); exit(0); }
    } 

    /* Now, we must find the indel_aa+num_indels as it is in the globals array. Continue using
     * the faithful value indel_aa..., and this time, while it is less than indel_aa+num_indels,   
     * continue incrementing it.  Start by setting num_indels_in_globals equal to position in
     * globals, and we will have a perfect for loop coming up.
     * If we are on exactly the same position as a previous insertion event, we want to do the
     * insertion after it.  Thus, we skip all of the spots of all of the previous insertions.
     * I am not going to put this just yet, though, because there is the possibility of an out
     * of order globals array, if instead of a non-ancestral sequence, there is a delete from an
     * ancestral sequence.  In this case, that position would be skipped, and it would make the
     * globals array disorganized. Thus, not yet including that situation.
     * 10-26-04:  I think that now we know that we are on a spot, we no longer have to perform the  
     * isSpot check.  This, of course, is only true for insertions (b/c every insertion will make
     * another spot, thus, don't need to check for it.  So, Only do the indel_aa_in_globals and
     * num_indels_in_globals thing, and should come out hunky-dory. */
    *num_indels_in_globals = *position_in_globals; 
    while(indel_aa_in_globals <= start_aa+indel_size) {
	indel_aa_in_globals++;
        /* Unfortunately, in the below for loop, there is no way that we can tell if there is a spot
         * or not.  In other words, if the first isSpot check does not find a spot, we will end up
         * pushing the same insert forward by one each time, and cannot determine if there has been
         * a wrong step or not.  So, either feed this num_skip in, or keep track of the number of 
         * isSpots falsely identified.  For this occasion, I think that a num_skip will work fine,
         * although I am not confident it will for the Delete case. */
        *num_indels_in_globals += 1; 
    } 
      
    if(insert_debug) fprintf(stderr,"  niig=%d\tpig=%d\n",*num_indels_in_globals,*position_in_globals); 

    /* In this function, we now have the bounds of the indels as they are represented in the  
     * globals array.  So, we now make the insertions into the globals array.  temp_in_globals  
     * is a construct that allows us to use isSpot to find each relevant insertion.  */ 
    for(i = *position_in_globals; i </*=*/ *num_indels_in_globals-num_skip; i++) {
        /* Make insertions BEFORE the amino acid [i]. First, move all of the spots in the 
         * global arrays down, so that there is room for the new insertion. */ 
        for(j = global_alignment_end; j >= i; j--) { 
            global_alignment_action[j+1] = global_alignment_action[j];
            global_alignment_fromAnc[j+1] = global_alignment_fromAnc[j];
        } 
        global_alignment_end+=1; 
        /* Now, make the insertions. */ 
        global_alignment_action[i] = 'i'; 
        global_alignment_fromAnc[i] = des->mytipNo; 
        //printf("i = %d\n",i); 
    } 
          
    //Print_Globals(); 

    for(i=final_aa; i>=start_aa+indel_size; i--) { /* Original: i>s_aa+i_size ... iPG >= */
	des->sequence[i]=des->sequence[i-indel_size];	/* Move the AA's to their position after insertion. */
	if(invariableSites) des->invariable[i]=des->invariable[i-indel_size];
    }
    if(invariableSites) for(i=start_aa; i<start_aa+indel_size; i++) { des->invariable[i]=0; }

    insertFillSequence(des->sequence,start_aa,indel_size);

    if(insert_debug) fprintf(stderr,"Leaving Insert, step %d\n",step++); // 6

    // This is error checking...
    /*fprintf(stderr,"Inserted Sequence (indel_size=%d):\t",indel_size);
    for(i=start_aa; i<start_aa+indel_size; i++) {
	fprintf(stderr,"%c",stateCharacters[(int)des->sequence[i]]);
    }
    fprintf(stderr,"\n\n");*/
    // When done, un-comment this stuff.
}

void Delete(TNode *des, int indel_size) {
    int start_aa;
    int final_aa;
    int i,j=0;
    int indel_aa_in_globals = 0; 
    int *position_in_globals; 
    int *num_indels_in_globals; 
    int *temp_in_globals; 
    int num_checks = 0; 
    int t_i_g_prev = 0;
    int num_rounds = 0; 
    int step = 0;

    final_aa=numSites+des->indels;
    start_aa=Find_Start_Location(des,&indel_size,numSites+des->indels,DELETE);

    if(delete_debug) fprintf(stderr,"DELETE: size = %d,\tspot = %d,\tfinal_aa = %d\tdes->tipNo = %d\n",indel_size,start_aa,final_aa,des->tipNo);
    if(delete_debug) fprintf(stderr,"DELETE: size = %d,\tfinal_aa-start_aa = %d\n",indel_size,final_aa-start_aa);

    position_in_globals = (int*)malloc(sizeof(int));
    *position_in_globals = 0;
    num_indels_in_globals = (int*)malloc(sizeof(int));
    *num_indels_in_globals = 0;
    temp_in_globals = (int*)malloc(sizeof(int));
     
    /* Here also, I need to find the correct spot in the global_* arrays, and place the
     * insertion symbol, as well as the TNode* tipNo. 
     * First, get to the indel_aa spot in the global arrays.  Because indel_aa is the position
     * before the insertion, we need to skip over indel_aa. */
    int prev_pos = 0; 
    while(indel_aa_in_globals <= start_aa) { 
        if(*position_in_globals > global_alignment_end) {
            printf("DELETE: Over g_a_end.  indel_aa_in_globals = %d,\tstart_aa = %d\tfinal_aa = %d\tdelete_size = %d\n", 
		    indel_aa_in_globals,start_aa,final_aa,indel_size);
            exit(0);
        }
        prev_pos = *position_in_globals;
        if(isSpot(position_in_globals,des)) { indel_aa_in_globals++; }
        *position_in_globals += 1;
        //if(prev_pos == *position_in_globals) { printf("prev_pos = p_i_g, %d\n",prev_pos); exit(0); }
    }

    /* Now, we must find the indel_aa+num_indels as it is in the globals array. Continue using
     * the faithful value indel_aa..., and this time, while it is less than indel_aa+num_indels,
     * continue incrementing it.  Start by setting num_indels_in_globals equal to position in
     * globals, and we will have a perfect for loop coming up. */
    *num_indels_in_globals = *position_in_globals;
    while(indel_aa_in_globals <= start_aa+indel_size) {
        /* Not sure how well this will work, but giving it a try...  Since deletes are different than
         * inserts (in that they happen only to corresponding indel values, and cannot happen consecutively,
         * We need to give a small bit of lee-way to the deletion spots.  In other words, for every delete
         * that happens, two spaces will be skipped.  Thus, I am adding the num_rounds variable, that keeps
         * track of the number of times two spots will be skipped.  I will then add the num_rounds value to
         * the for loop end.  Thus, the true number of deletes should be added before ending the for loop */
        if(isSpot(num_indels_in_globals,des)) { indel_aa_in_globals++; num_rounds++; }
        *num_indels_in_globals += 1; 
    } 

    if(delete_debug) fprintf(stderr,"  niig=%d\tpig=%d\n",*num_indels_in_globals,*position_in_globals);

    /* In this function, we now have the bounds of the indels as they are represented in the
     * globals array.  So, we now make the insertions into the globals array.  temp_in_globals
     * is a construct that allows us to use isSpot to find each relevant insertion.  */ 
    *temp_in_globals = *position_in_globals;
    t_i_g_prev = *temp_in_globals;
    for(i = *position_in_globals; i </*=*/ *num_indels_in_globals+num_rounds; i++) { 
        *temp_in_globals = i;  // + num_checks; 
        if(isSpot(temp_in_globals,des)) { 
            /* Make deletions BEFORE the amino acid [i]. First, move all of the spots in the
             * global arrays down, so that there is room for the new deletion. */
            for(j = global_alignment_end; j >= /*>= b/c putting before*/ i; j--) {
                global_alignment_action[j+1] = global_alignment_action[j];
                global_alignment_fromAnc[j+1] = global_alignment_fromAnc[j];
            } 
            global_alignment_end+=1;
            /* Now, make the deletions. */ 
            global_alignment_action[i] = 'd';
            global_alignment_fromAnc[i] = des->mytipNo; 
            *temp_in_globals += 1;
        }    
        i = *temp_in_globals;
        //printf("i %d\t\n",i);
    }
     

    if(delete_debug) Print_Globals(); 
         
    j=0; 
    if(delete_debug) fprintf(stderr,"Deleted Sequence:\t");
    for(i=start_aa+indel_size; i<final_aa; i++) { 
        if(j<indel_size) { 
	    if(delete_debug) fprintf(stderr,"%c",stateCharacters[(int)des->sequence[i-indel_size]]); 
	    j++; 
	}
        des->sequence[i-indel_size]=des->sequence[i];
        if(invariableSites) des->invariable[i-indel_size]=des->invariable[i]; 
    } 
    if(delete_debug) fprintf(stderr,"\n\n"); 
} 
                 
              
/* This function finds the start location for the insertion or deletion action, taking each action as a special
 * case. */ 
int Find_Start_Location(TNode *des, int *indel_size, int final_aa, int action) {
    int location_in_sequence=0; 
    short *acceptable_positions=(short*)malloc((final_aa+2)*sizeof(short));
    int position;
    int i,j; 
    int FSL_debug = 0;
    int ok; 
    int numAcceptablePositions;
    int step = 0;

    if(FSL_debug) fprintf(stderr,"Find_Start_Location, step %d\n",step++);  // 0

    for(i=0; i<=final_aa; i++) { acceptable_positions[i]=0; }

    if(action == INSERT) {
        numAcceptablePositions=final_aa;
        for(i=0; i<=final_aa; i++) {
            acceptable_positions[i] = 1;        /* Set to acceptable to begin with. */
            if(invariableSites) { 
                if((des->invariable[i]==NO_INDEL || des->invariable[i]==INVAR_AND_NOINDEL) && 
                   (des->invariable[i-1]==NO_INDEL || des->invariable[i-1]==INVAR_AND_NOINDEL)) {
                    acceptable_positions[i]=0; 
                    numAcceptablePositions--;
                } 
            }
        }        
        acceptable_positions[final_aa]=1;
    } else { 
	ok = 0;
	while(!ok) {
	    /* A deletion cannot occur past the deletion size.  Otherwise, it will lop off more than the end of the seq. */
            numAcceptablePositions=final_aa-*indel_size;
	    if(FSL_debug) fprintf(stderr,"numAcceptablePositions = %d\tfinal_aa = %d\tindel_size = %d\n",
				  numAcceptablePositions,final_aa,*indel_size);
            for(i=0; i<=final_aa-*indel_size; i++) { 
                acceptable_positions[i] = 1;
                if(invariableSites) { 
                    for(j=i; j<i+*indel_size; j++) {
                        if(des->invariable[j]) {
                            acceptable_positions[i]=0;
                            numAcceptablePositions--;
			    // Need to short-circuit this loop, or we will continue to decrement the numAccPos too far.
			    j=i+*indel_size;
                        }
                    }  
                } 
            }  
	    /* On large deletes, if there are invariable positions, it is quite possible that there will not be
	     * a position such that a delete can fit.  Thus, in this loop, we will decrement the indel size by
	     * 1, and try again. */
    	    if(numAcceptablePositions < 0) *indel_size = *indel_size-1;
    	    else ok = 1;
	}
    } 

    if(numAcceptablePositions < 0) {
	fprintf(stderr,"evolve.c, Find_Start_Locations:\nNegative number of acceptable positions.\n");
	fprintf(stderr,"numAcceptablePositions = %d\n",numAcceptablePositions);
	exit(0);
    }
    if(FSL_debug) fprintf(stderr,"Find_Start_Location, step %d\tnumAcceptablePositions = %d\n",
			  step++,numAcceptablePositions);  // 1

    if(FSL_debug) {
    	if(action == DELETE) {
    	    fprintf(stderr,"acceptable positions for %s:",((action)?"deletion":"insertion")); 
    	    for(i=0; i<final_aa; i++) { 
            	fprintf(stderr,"%d",acceptable_positions[i]);
    	    } fprintf(stderr,"\n"); 
    	    if(invariableSites && indel_flag) {
            	fprintf(stderr,"Invariable sites for     %s:",((action)?"deletion":"insertion"));
            	for(i=0; i<final_aa; i++) {
            	    fprintf(stderr,"%d",des->invariable[i]);
            	} fprintf(stderr,"\n");
    	    } 
    	}
        fprintf(stderr,"Find_Start_Location, step %d\n",step++);  // 2
    }

    /* If the invariable array is ALL 1's, then there can be no action. */
    ok=0; /* Use this variable to find if there is at least one acceptable position to make indel */
    i=0; 
    while(!ok && i</*=*/final_aa) { 
        if(acceptable_positions[i]) { ok=1; } 
        i++; 
    } 

    if(FSL_debug) fprintf(stderr,"Find_Start_Location, step %d\n",step++);  // 3

    if(i >= final_aa) {
	//  Find out whether the acceptable positions or 
	fprintf(stderr,"numSites = %d\ndes->indels = %d\nfinal_aa = %d\n",numSites,des->indels,final_aa);
	fprintf(stderr,"invariable array (top), accepting positions (bottom):\n");
	for(ok=0; ok < final_aa; ok++) { fprintf(stderr,"%d",des->invariable[ok]); }
	fprintf(stderr,"\n");
	for(ok=0; ok < final_aa; ok++) { fprintf(stderr,"%d",acceptable_positions[ok]); }
        fprintf(stderr,"There is no acceptable position to make %s of size %d.\n",
	        ((action)?"deletion":"insertion"),*indel_size);
        exit(0);
    } 
             
    if(FSL_debug) fprintf(stderr,"numAcceptablePositions = %d\t",numAcceptablePositions);
    i=(int)(rndu()*(double)numAcceptablePositions); 
    if(FSL_debug) fprintf(stderr,"i = %d\n",i);
    position=0; j=0;
    while(!acceptable_positions[j]) j++;        // Get to the first 1 in a_p, or else an i=0 will always delete first pos.
    while(position < i && j </*=*/ final_aa) {
        if(acceptable_positions[j]) position++;
        if(position < i) j++;
	if(FSL_debug) fprintf(stderr,"position = %d\ti = %d\tj = %d\n",position,i,j);
    } 
         
    if(FSL_debug) fprintf(stderr,"Find_Start_Location, step %d\n",step++);  // 4
    if(FSL_debug) fprintf(stderr,"The %dth acceptable indel position is the %dth spot.\n",position,j);
             
    free(acceptable_positions);  
    return j; 
} 
                     
int Find_Indel_Size(double *dist, int max_indel) {
    int size=0;   
    int i=1;  
    double sum=0;
    double rand; 
     
    rand=rndu(); 
    if(rand==0) rand=rand+0.00001; 
     
    while(i <= max_indel && size == 0) { 
        sum+=dist[i];
        if(sum > rand) size=i; 
        i++; 
    } 
         
    return size;
}        
     
int  Find_Action(double len, double P_insert_, double P_delete_) {
    double P_indel_; 
    double indel_test_value;
  
    if(P_insert_==0 && P_delete_==0) {
        P_indel_ = 0.0224 - 0.0219 * exp(-0.01168*(len*100));
	if((double)rndu() < P_indel_) {
            if((double)rndu() < 0.5) return INSERT;
            else return DELETE;
        } else return NO_ACTION;
    } else {
        P_indel_=(P_insert_+P_delete_)*len;
        if((double)rndu() < P_indel_) {
       	    indel_test_value=P_delete_/(P_insert_+P_delete_);
	    if((double)rndu() < indel_test_value) return DELETE;
            else return INSERT;
        } else return NO_ACTION;
    } 
} 
         
/***** 
 * CS: This routine created to search for the ancestors of the TNode passed into it.  It is
 * a recursive function, that returns 1 if the ancestor does exist in the TNode's history,
 * and -1 if it does not.  This function is critical for the multiple alignment at the end
 * of the run of indel_Pseq-Gen.  */
int  isAnc(TNode *thisNode, int anc) {
    if(thisNode -> mytipNo == anc) {
        /* Found the ancestor, return 1. */
        return 1; 
    } else if(thisNode -> mytipNo == -1) {
        /* Then this is the root.  No need to check for  more ancestors. */
        return 0; 
    } else { 
        return isAnc(thisNode->anc, anc);
    } 
} 
             
/***** 
 * CS: This checks the global arrays, and if an insert is should be counted for a certain taxa,
 * then this function returns a 1.  If not, then it returns a 0.  Method explained in function.
 *****/
int  isSpot(int *pos, TNode *curr_taxa) {
    int new_pos;
 
    /* Find the first insert in the array, don't care which taxa made it. */ 
    while(global_alignment_action[*pos] != 'i' && *pos < global_alignment_end) { *pos+=1; }
  
    if(!isAnc(curr_taxa,global_alignment_fromAnc[*pos])) { return 0; }
    else { 
        /* the if stmt determines that this node is our ancestor, so now we need to dig around
         * for the deletes that may have happened to this insertion.  If we cannot find any
         * one spot where a delete happened in our ancestry to this particular insertion, then
         * we can say that this is a true spot in our protein sequence. */
        /* set our position pointer to be behind the curr insertion. Deletes happen here */
        new_pos = *pos-1;
        while(new_pos > 0 && global_alignment_action[new_pos] != 'i') { new_pos--; }
         
        /* We are now at the previous insertion stmt.  Go one step ahead. */
        new_pos++;
             
        /* If the new_pos is equal to the old pos, then there are no deletes, and we have found
         * that the spot is valid. */ 
        if(new_pos == *pos) { return 1; } 
        else {
            /* Now we know that there are some delete statements.  Want to find if these delete
             * statements happened in the current taxa's history, or a different taxa's history. */
            while(global_alignment_action[new_pos] == 'd') {
                /* If it is a delete, and it is an ancestor, then this insert IS NOT a countable spot. */
                if(isAnc(curr_taxa,global_alignment_fromAnc[new_pos])) { return 0; } 
                new_pos++;
            } 
            /* If the while loop does not do a return of -1, then it must be a spot, therefore ... */
            return 1; 
        } 
    } 
} 
         
void EvolveNode(TTree *tree, TNode *anc, TNode *des, int inFromSite, int inNumSites, double scale, int *P)
{ 
    double len;   
    int i;   
    int EvolveNode_debug = 0;
    int step=0;

    if(EvolveNode_debug) fprintf(stderr,"EvolveNode, step %d\n",step++); // 0

    if(indel_flag && invariableSites) {
        Create_indelRates(des); 
        for(i=0; i<inNumSites+tree_indel_buffer; i++) {
            des->invariable[i]=anc->invariable[i]; 
        } 
    } 

    if(EvolveNode_debug) fprintf(stderr,"EvolveNode, step %d\n",step++); // 1

    len=des->length0*scale; 
                 
    des->indels=anc->indels;    /*CS: Pass # of indels of anc to des.*/
    des->anc=anc;               /*CS: Pointer, equivalent to *branch0 except for "root" in unrooted tree. */

    if(EvolveNode_debug) fprintf(stderr,"EvolveNode, step %d\tinNumSites=%d\tdes->indels=%d\tancestor=%d\n",step++,inNumSites,des->indels,anc->mytipNo); // 2

    // Since we are no longer passing des->indels into the function, memcpy needs to be changed to reflect it.
    memcpy(des->sequence, anc->sequence, inNumSites+des->indels);

    if(EvolveNode_debug) fprintf(stderr,"EvolveNode, step %d\n",step++); // 3

    // 012606: Same issue.  Wherever
    MutateSequence(des, des->sequence, inFromSite, inNumSites+des->indels, len);

    if(EvolveNode_debug) fprintf(stderr,"EvolveNode, step %d\n",step++); // 4

    *P+=1;   
    des->mytipNo=*P;
         
    /*CS: This function implements the indel routines: Need to pass TTree *tree because of all of the indel
     * information carried in the tree, such as P_ins_ and P_del_. */
    if(EvolveNode_debug) fprintf(stderr,"EvolveNode: Before InDels, for mytipNo=%d\n",des->mytipNo);

    if (tree->do_indels) InDel(tree,des,len);

    if(EvolveNode_debug) fprintf(stderr,"EvolveNode: After InDels, for mytipNo=%d\n",des->mytipNo);
     
    /* Don't need to pass the indels in these calls, since the indels are kept separately. */
    /* 012606: By passing the indels, inNumSites keep reducing.  Ex - if a delete is made early in the tree,
     * the next pass of EvolveNode will be called with inNumSites+des->indels, which will be a smaller number
     * than inNumSites.  However, on the next call, the EvolveNode will again be called with a smaller number,
     * even if no insertions or deletions were made.  So, sucks to pass des->indels! */
    if (des->tipNo==-1) { 
        EvolveNode(tree, des, des->branch1, inFromSite, inNumSites/*+des->indels*/, scale, P);
        EvolveNode(tree, des, des->branch2, inFromSite, inNumSites/*+des->indels*/, scale, P);
    } 
} 
      
                 
void EvolveSequences(TTree *tree, int inFromSite, int inNumSites, double scale, char *ancestor)
{                
    int i,j; 
    int *node_num; 
    FILE        *rootSeq;
     
    //fprintf(stderr,"EvolveSeq: %d %d %d\n",indelRoot,invariableSites,indel_flag);
     
    node_num= (int*)malloc(sizeof(int));
    *node_num= -1;  
         
    Create_Global_Arrays(tree); 
      
    tree->root->indels=0; 
     
    if (indel_flag && invariableSites) { 
        Create_indelRates(tree->root);
    } 
         
    if (indelRoot) {
        SetSequence(tree->root->sequence, indelRootSequence ,inFromSite, inNumSites);
        if (invariableSites) {
            if(indel_flag) {
                for(i=0; i<inNumSites+tree_indel_buffer; i++) { 
                    if(i>inNumSites) tree->root->invariable[i]=0;
                    else tree->root->invariable[i]=indelRootInvariable[i];
                }  
            } else { 
                for(i=0; i<inNumSites; i++) invariable[i]=indelRootInvariable[i];
            } 
        }
    } else if (ancestor==NULL) RandomSequence(tree->root->sequence, inNumSites);
    else SetSequence(tree->root->sequence, ancestor, inFromSite, inNumSites);
         
    if((rootSeq = fopen("root.seq-gen","w")) == NULL) {
        printf("Could not open \"root.seq-gen\" to write.\n");
        exit(0); 
    }
    for (j=0; j<numSites; j++) fputc(stateCharacters[tree->root->sequence[j]],rootSeq);
    fputc('\n', rootSeq); 
    fclose(rootSeq);
         
    if(!indelRoot && invariableSites && indel_flag) {
        for (i=0; i<numSites; i++) tree->root->invariable[i] = IsInvariable(); 
        for (i=numSites; i<numSites+tree_indel_buffer; i++) tree->root->invariable[i]=0;
    } else if(invariableSites && !indelRoot) {
        for(i=0; i<numSites; i++) invariable[i]=IsInvariable(); 
    } 
                     
    /* Rescale the branch lengths to accommodate the invariable sites */
    if (invariableSites) scale /= (1.0 - proportionInvariable);
                 
    if (invariableSites) SetSiteRates(tree->root,inFromSite, inNumSites, scale);
         
    EvolveNode(tree, tree->root, tree->root->branch1, inFromSite, inNumSites, scale, node_num);
    EvolveNode(tree, tree->root, tree->root->branch2, inFromSite, inNumSites, scale, node_num);
    if (!tree->rooted)
        EvolveNode(tree, tree->root, tree->root->branch0, inFromSite, inNumSites, scale, node_num);
} 
         
void WriteSequences(FILE *fv, int treeNo, int datasetNo, TTree **treeSet, int *partitionLengths)
{ 
     
    switch (fileFormat) {
    case PHYLIPFormat:
    case RelaxedFormat: 
        WritePhylipFormat(fv, treeSet, partitionLengths); 
        break; 
    case NEXUSFormat: 
        WriteNexusFormat(fv, treeNo, datasetNo, treeSet, partitionLengths);
        break;
    }                
} 
     
void WritePhylipFormat(FILE *fv, TTree **treeSet, int *partitionLengths)
{ 
    int i, j, k;
    char *P; 
     
    fprintf(fv, " %d %d\n", numTaxa, numSites);
         
    for (i=0; i<numTaxa; i++) {
        if (fileFormat == RelaxedFormat) { fprintf(fv, "%s ", treeSet[0]->names[i]);
            fprintf(stderr, "%s ", treeSet[0]->names[i]); 
        } else {
            j=0;
            P=treeSet[0]->names[i];
            while (j<10 && *P) {
                fputc(*P, fv);
                j++; 
                P++;
            } 
            while (j<10) { 
                fputc(' ', fv);
                j++; 
            }
        }
        for (k = 0; k < numPartitions; k++) { 
            P=treeSet[k]->tips[i]->sequence;
            for (j=0; j<partitionLengths[k]+treeSet[k]->tips[i]->indels; j++) {
                fputc(stateCharacters[(int)*P], fv);
                P++;
            } 
        }
        fputc('\n', fv); 
    } 
} 
         
void WriteNexusFormat(FILE *fv, int treeNo, int datasetNo, TTree **treeSet, int *partitionLengths)
{ 
        int i, j, k, len, maxLen;
        char *P; 
                 
        if (treeNo > 0 && datasetNo > 0)
                fprintf(fv, "Begin DATA;\t[Tree %d, Dataset %d]\n", treeNo, datasetNo);
        else if (treeNo > 0)
                fprintf(fv, "Begin DATA;\t[Tree %d]\n", treeNo);
        else if (datasetNo > 0)
                fprintf(fv, "Begin DATA;\t[Dataset %d]\n", datasetNo);
        else
                fprintf(fv, "Begin DATA;\n"); 
             
        fprintf(fv, "\tDimensions NTAX=%d NCHAR=%d;\n", numTaxa, numSites); 
        if (isNucModel) { 
                fprintf(fv, "\tFormat MISSING=? GAP=- DATATYPE=DNA;\n");
        } else {
                fprintf(fv, "\tFormat MISSING=? GAP=- DATATYPE=PROTEIN;\n");
        } 
      
        fprintf(fv, "\tMatrix\n");
         
        maxLen = strlen(treeSet[0]->names[0]); 
        for (i=1; i<numTaxa; i++) {
                len = strlen(treeSet[0]->names[i]);
                if (len > maxLen)
                        maxLen = len;
        } 
                 
        for (i=0; i<numTaxa; i++) {
                fprintf(fv, "%s ", treeSet[0]->names[i]); 
                len = maxLen - strlen(treeSet[0]->names[i]);
                for (j = 0; j < len; j++) { 
                        fputc(' ', fv);
                } 
             
                for (k = 0; k < numPartitions; k++) { 
                        P=treeSet[k]->tips[i]->sequence;
                        for (j=0; j<partitionLengths[k]; j++) { 
                                fputc(stateCharacters[(int)*P], fv);
                                P++; 
                        }
                }
                fputc('\n', fv);  
        }
        fprintf(fv, "\t;\nEND;\n\n"); 
} 
                 
void WriteAncestralSequences(FILE *fv, TTree *tree)
{ 
    int j, n;
    char *P;     
         
    if (!tree->rooted) n = (2 * numTaxa) - 3; 
    else n = (2 * numTaxa) - 2; 
    fprintf(fv, " %d %d\n", n, numSites);   
                         
    n = numTaxa + 1;
             
    fprintf(fv, "%d\t", n); 
    P=tree->root->sequence; 
    for (j=0; j<numSites; j++) { 
        fputc(stateCharacters[(int)*P], fv); 
        P++; 
    } 
    fputc('\n', fv);
                 
    if (!tree->rooted) WriteAncestralSequencesNode(fv, tree, &n, tree->root->branch0);
    WriteAncestralSequencesNode(fv, tree, &n, tree->root->branch1);
    WriteAncestralSequencesNode(fv, tree, &n, tree->root->branch2);
}                
 
void WriteAncestralSequencesNode(FILE *fv, TTree *tree, int *nodeNo, TNode *des)
{ 
    int j;       
    char *P = des->sequence;
     
    if (des->tipNo==-1) { 
        (*nodeNo)++; 
                         
        fprintf(fv, "%d\t", *nodeNo);
        for (j=0; j<numSites; j++) {
            fputc(stateCharacters[(int)*P], fv);
            P++; 
        } 
        fputc('\n', fv); 
         
        WriteAncestralSequencesNode(fv, tree, nodeNo, des->branch1);
        WriteAncestralSequencesNode(fv, tree, nodeNo, des->branch2);
    } else {     
        fprintf(fv, "%s\t", tree->names[des->tipNo]); 
        for (j=0; j<numSites; j++) { 
            fputc(stateCharacters[(int)*P], fv); 
            P++; 
        }
        fputc('\n', fv); 
    }
     
} 
     
     
void WriteRates(FILE *fv)
{                        
        int i; 
         
        fprintf(fv, "Relative rates for each site:\n");
             
        for (i=0; i<numSites; i++) {
                fprintf(fv, "%d\t%lf\n", i+1, siteRates[i]);
        }
        fprintf(fv, "\n"); 
} 
     
         
void Print_Globals() { 
    int i; 
    int temp;    
         
    for(i=1; i<global_alignment_end; i++) {
        if(global_alignment_fromAnc[i] == -1) { fprintf(stderr,"-"); }
        else {
            if(global_alignment_fromAnc[i] > 9) {
                temp = global_alignment_fromAnc[i];
                while(temp > 9) {
                    temp/=10;
                    fprintf(stderr," ");
                }
            }
            fprintf(stderr,"%c",global_alignment_action[i]);
        }    
    } 
    fprintf(stderr,"\n"); 
    for(i=1; i<=global_alignment_end; i++) {
        if(global_alignment_fromAnc[i] == -1) { fprintf(stderr,"-"); }
        else { fprintf(stderr,"%d",global_alignment_fromAnc[i]); }
    }
    fprintf(stderr,"\n");
} 
     
void WriteMA(FILE *fv, TTree *tree)
{        
    int i, j, k, end_loc, keep, i2, max_i2;
    char **print; 
    int  *invar_print;
    int  *taxa_ptrs; 
    int  deletion, column; 
    char *P; 
    FILE        *MA; 
                     
    print = (char**)malloc((tree->numTips+1)*sizeof(char*));
    taxa_ptrs = (int*)malloc((tree->numTips+1)*sizeof(int));
    /* Create and use invar_print, even if there are not invar sites.  However, initialize them to '0' to reflect that
     * there are not invariable sites in the run.  This allows indel-Seq-Gen to be easier coding. */
    invar_print = (int*)malloc(10000*sizeof(int));
    for(i=0; i<=tree->numTips; i++) {
        taxa_ptrs[i] = 0; 
        print[i] = (char*)malloc(10000*sizeof(char)); 
        for(j=0; j<10000; j++) { 
            print[i][j] = '\0';
            invar_print[j]=0;
        }
    }
 
    /* To write out the file, I am going to work from the beginning of the alignment to
     * the end of the alignment, making a decision for each sequence whether to output an
     * amino acid or a '-', in each position.  Each decision will be stored in the array
     * called print.  Also, will keep an array of pointers to positions in each of the
     * taxa arrays.  */
    if(global_alignment_action[0] == 'd') { i=0; } else { i=1; }
    i2=0; max_i2=0;
    while(i < global_alignment_end) {
        /* Absolutely necessary to begin on an insertion on each pass, since 'i' is the only
         * case when an action is deemed necessary.  'd' is only an 'i' modifier. */
        while(global_alignment_action[i] != 'i' && i <= global_alignment_end) { i++; }
     
        for(j=0; j<tree->numTips; j++) { 
            /* First, and easiest, case.  If the AA is not in your ancestry, output '-'.  There
             * are no special cases to this particular instance. */
            if(!isAnc(tree->tips[j],global_alignment_fromAnc[i])) {
                print[j][i2] = '-'; 
            } else { 
                /* Another easy case.  If there are no deletes, then just transfer all of the
                 * amino acids to the print array.  Since we know that the insert is an ancestor,
                 * no need to make any other checks than looking at the previous action. */
                if(global_alignment_action[i-1] == 'i') {
                    print[j][i2] = tree->tips[j]->sequence[taxa_ptrs[j]];
                    if(invariableSites) invar_print[i2] = tree->tips[j]->invariable[taxa_ptrs[j]];
                    taxa_ptrs[j]++; 
                /* Now, all the easy work is done.  Need to now go backwards to find the next insert,
                 * and from that insertion, discover all of the deletions that have happened to the
                 * current position.  Then, from those deletions, need to output a '-' only if the
                 * deletion happened at an ancestral sequence.  Otherwise, output our happy little
                 * amino acid.  */
                } else { 
                    /* This can still be an insertion.  If so, i = keep after the next comment stmts. */
                    keep = i; 
                    /* The following three actions place our attention at the first delete statement. */
                    i--;
                    while(global_alignment_action[i] != 'i' && i >= 0) { i--; }
                    i++; 
                    if(keep == i) { 
                        if(global_alignment_fromAnc[i] == tree->tips[j]->tipNo) {
                            print[j][i2] = tree->tips[j]->sequence[taxa_ptrs[j]];
                            if(invariableSites) invar_print[i2] = tree->tips[j]->invariable[taxa_ptrs[j]];
                            taxa_ptrs[j]++; 
                        } else { print[j][i2] = '-'; } 
                    } else { 
                        /* Now, on the delete statement.  Find if it is an ancestral delete.  Need to check,
                         * this for all deletes, so will do this with a while statement. */
                        deletion = 0; /* Begin by assuming that the deletion is not in the sequence. */
                        while(global_alignment_action[i] != 'i' && i <= global_alignment_end) {
                            if(isAnc(tree->tips[j],global_alignment_fromAnc[i])) { 
                                deletion = 1; 
                            } 
                            i++; 
                        } 
                        /* Now, we have found where the deletion is, so now just perform action. */
                        if(deletion) { print[j][i2] = '-'; }
                        else { 
                            print[j][i2] = tree->tips[j]->sequence[taxa_ptrs[j]];
			    if(invariableSites) invar_print[i2] = tree->tips[j]->invariable[taxa_ptrs[j]];
                            taxa_ptrs[j]++; 
                        }
                    } 
                } 
            } 
        } 
        i++; 
        i2++; 
        if(i2 > max_i2) { max_i2 = i2; } 
    } 
                         
    for(j=0; j<max_i2; j++) { 
        i=0; 
        column = 1;  /* Default assume a column is all '-' */ 
        while(i < tree->numTips && column) { 
            if(print[i][j] != '-') { column = 0; }
            i++; 
        } 
        if(column) { 
            max_i2--;
            j--;
            //printf("!!\n\n");
            for(i=j+1; i<max_i2; i++) { 
                for(k=0; k<tree->numTips; k++) {
                    print[k][i] = print[k][i+1];
                } 
            } 
        } 
    } 

    /* Opening a file for the multiple alignment to go into. */
    if((MA = fopen("iSG_MA","w")) == NULL) {
        printf("Could not open \".mult_align.iPG\" to write.\n");
        exit(0); 
    } 

    /* Making MA file as similar as possible to the other output file (for indel-Seq-Gen).  Thus, a top line is needed */
    for(i=0; i<max_i2; i++) {
        fprintf(MA, "%d",invar_print[i]);
    }
    fprintf(MA,"\n");
 
    /* Print the information in the print array. */
    for(i=0; i<tree->numTips; i++) {
        if(1) { 
            if (1) {
                fprintf(MA, "%s ", tree->names[i]);
            } else {
                j=0;
                P=tree->names[i];
                while (j<10 && *P) {
                    fputc(*P, MA); 
                    j++; 
                    P++; 
                }
                while (j<10) {
                    fputc(' ', MA);
                    j++; 
                } 
            } 
            for(j=0; j<max_i2; j++) {
                if(print[i][j] == '-') { 
		    fputc('-', MA); 
                } else { 
		    fprintf(MA, "%c", stateCharacters[print[i][j]]); 
		}
            }
            fputc('\n', MA);
        } 
    } 
    fclose(MA); 
                     
    return; 
} 
